"""Tests for annotate."""

from __future__ import annotations

import sys
from io import StringIO
from tokenize import generate_tokens
from typing import TYPE_CHECKING

import pytest

from idlemypyextension import annotate

if TYPE_CHECKING:
    from collections.abc import Collection, Sequence


def test_parse_error() -> None:
    with pytest.raises(annotate.ParseError, match=""):
        raise annotate.ParseError()


@pytest.mark.parametrize(
    ("text", "expect"),
    [("  waf", 2), ("cat", 0), ("     fish", 5), ("   ", 3), ("", 0)],
)
def test_get_line_indent(text: str, expect: int) -> None:
    assert annotate.get_line_indent(text, " ") == expect


@pytest.mark.parametrize(
    ("name", "result"),
    [
        ("typing.Set", "set"),
        ("typing.TypedDict", "TypedDict"),
        ("mypy_extensions.KwArg", "KwArg"),
        ("builtins.list", "builtins.list"),
    ],
)
def test_get_type_repr(name: str, result: str) -> None:
    assert annotate.get_type_repr(name) == result


@pytest.mark.parametrize(
    ("typevalue", "result"),
    [
        (annotate.TypeValue("typing.Set"), "set"),
        (annotate.TypeValue("typing.TypedDict"), "TypedDict"),
        (annotate.TypeValue("mypy_extensions.KwArg"), "KwArg"),
        (annotate.TypeValue("builtins.list"), "builtins.list"),
        (annotate.TypeValue("List"), "list"),
        (
            annotate.TypeValue(
                "List",
                (annotate.TypeValue("list", (annotate.TypeValue("int"),)),),
            ),
            "list[list[int]]",
        ),
        (
            annotate.TypeValue(
                "Union",
                (annotate.TypeValue("str"), annotate.TypeValue("None")),
            ),
            "str | None",
        ),
        (
            annotate.TypeValue(
                "Union",
                (
                    annotate.TypeValue("set", (annotate.TypeValue("int"),)),
                    annotate.TypeValue("None"),
                ),
            ),
            "set[int] | None",
        ),
    ],
)
def test_get_typevalue_repr(
    typevalue: annotate.TypeValue,
    result: str,
) -> None:
    assert annotate.get_typevalue_repr(typevalue) == result


def test_typevalue_repr() -> None:
    assert (
        repr(
            annotate.TypeValue(
                "Union",
                (
                    annotate.TypeValue("set", (annotate.TypeValue("int"),)),
                    annotate.TypeValue("None"),
                ),
            ),
        )
        == "TypeValue('Union', (TypeValue('set', (TypeValue('int'),)), TypeValue('None')))"
    )


def test_typevalue_eq() -> None:
    assert annotate.TypeValue(
        "Union",
        (
            annotate.TypeValue("set", (annotate.TypeValue("int"),)),
            annotate.TypeValue("None"),
        ),
    ) == annotate.TypeValue(
        "Union",
        (
            annotate.TypeValue("set", (annotate.TypeValue("int"),)),
            annotate.TypeValue("None"),
        ),
    )


def test_typevalue_super_eq() -> None:
    assert (
        annotate.TypeValue(
            "Union",
            (
                annotate.TypeValue("set", (annotate.TypeValue("int"),)),
                annotate.TypeValue("None"),
            ),
        )
        != "potato"
    )


@pytest.mark.parametrize(
    ("items", "result"),
    [
        (("waffles",), "waffles"),
        (("waffles", "pancakes"), "waffles or pancakes"),
        (
            ("waffles", "pancakes", "a muffin"),
            "waffles, pancakes, or a muffin",
        ),
        (
            ("waffles", "pancakes", "cereal", "a muffin"),
            "waffles, pancakes, cereal, or a muffin",
        ),
    ],
)
def test_list_or(items: Collection[str], result: str) -> None:
    assert annotate.list_or(items) == result


@pytest.mark.parametrize(
    ("text", "tokens"),
    [
        (
            "(int, float) -> str",
            [
                annotate.Separator("("),
                annotate.DottedName("int"),
                annotate.Separator(","),
                annotate.DottedName("float"),
                annotate.Separator(")"),
                annotate.Separator("->"),
                annotate.DottedName("str"),
                annotate.End(),
            ],
        ),
        (
            "pytz.tzfile.America/Los_Angeles",
            [
                annotate.DottedName("datetime.tzinfo"),
                annotate.End(),
            ],
        ),
    ],
)
def test_tokenize(text: str, tokens: list[annotate.Token]) -> None:
    result = annotate.tokenize(text)
    if result != tokens:
        print(f"{tokens}\n!=\n{result}")
    assert result == tokens


def test_invalid_tokenize() -> None:
    with pytest.raises(
        annotate.ParseError,
        match="Could not parse '\\$HOME' from 'path.\\$HOME'",
    ):
        annotate.tokenize("path.$HOME")


@pytest.mark.parametrize(
    ("text", "tokens"),
    [
        (
            "def test(value: int = -~ALL_BITS):",
            [
                annotate.Definition("def"),
                annotate.FunctionName("test"),
                annotate.EndSeparator("("),
                annotate.ArgumentName("value"),
                annotate.TypeDef(":"),
                annotate.DottedName("int"),
                annotate.DefaultDef("="),
                annotate.ArgumentDefault("-~ALL_BITS"),
                annotate.EndSeparator(")"),
                annotate.EndDefinition(":"),
                annotate.End(),
            ],
        ),
        (
            """def potato(
    get_line = lambda lno: GLOBAL_LINES[lno]
):""",
            [
                annotate.Definition("def"),
                annotate.FunctionName("potato"),
                annotate.EndSeparator("("),
                annotate.EndSeparator(text="\n"),
                annotate.EndSeparator(text="    "),
                annotate.ArgumentName("get_line"),
                annotate.DefaultDef("="),
                annotate.ArgumentDefault(text="lambda"),
                annotate.LambdaBody("lno: GLOBAL_LINES[lno]"),
                annotate.EndSeparator(text="\n"),
                annotate.EndSeparator(text=""),
                annotate.EndSeparator(")"),
                annotate.EndDefinition(":"),
                annotate.End(),
            ],
        ),
    ],
)
def test_tokenize_definition(text: str, tokens: list[annotate.Token]) -> None:
    lines = text.splitlines(True)

    def get_line(line_no: int) -> str:
        return lines[line_no]

    result, _ = annotate.tokenize_definition(0, get_line)

    if result != tokens:
        print(f"{tokens}\n!=\n{result}")
    assert result == tokens


def test_invalid_tokenize_definition() -> None:
    lines = "def def".splitlines(True)

    def get_line(line_no: int) -> str:
        return lines[line_no]

    with pytest.raises(
        annotate.ParseError,
        match="Did not expect second definition keyword",
    ):
        annotate.tokenize_definition(0, get_line)


@pytest.mark.parametrize(
    ("function_text", "arg_types", "return_type", "result", "filename"),
    [
        (
            """def format_id():""",
            [],
            "str",
            """def format_id() -> str:""",
            None,
        ),
        (
            """def format_id(user):""",
            ["int"],
            "str",
            """def format_id(user: int) -> str:""",
            None,
        ),
        (
            """def format_name(name = "TEMP"):""",
            ["str"],
            "str",
            """def format_name(name: str = "TEMP") -> str:""",
            None,
        ),
        (
            """def sleep(time = 1 / 3):""",
            ["float"],
            "None",
            """def sleep(time: float = 1 / 3) -> None:""",
            None,
        ),
        (
            """def format_id(
    user# format function inline comment
):""",
            ["int"],
            "str",
            """def format_id(
    user: int  # format function inline comment
) -> str:""",
            None,
        ),
        (
            """def foo(x: int, longer_name: str) -> None:""",
            ["int", "str"],
            "None",
            """def foo(x: int, longer_name: str) -> None:""",
            None,
        ),
        (
            """def waffle_magic(
    self, obj: wafflemodule.Waffle, action: str
) -> None:""",
            ["wafflemodule.Waffle", "str"],
            "None",
            """def waffle_magic(
    self, obj: wafflemodule.Waffle, action: str
) -> None:""",
            None,
        ),
        (
            """def waffle_magic(
    self, obj, action, sand
):""",
            ["Union[int, float, complex, wafflemodule.Waffle]", "str", "int"],
            "None",
            """def waffle_magic(
    self, obj: int | float | complex | wafflemodule.Waffle, action: str, sand: int
) -> None:""",
            None,
        ),
        (
            """def get_annotation(
    annotation,
    get_line
):""",
            ["dict[str, Any]", "Callable[[int], str]"],
            "tuple[str, int]",
            """def get_annotation(
    annotation: dict[str, Any],
    get_line: Callable[[int], str]
) -> tuple[str, int]:""",
            None,
        ),
        (
            """def potato(
    get_line = lambda lno: GLOBAL_LINES[lno]
):""",
            ["Callable[[int], str]"],
            "bool",
            """def potato(
    get_line: Callable[[int], str] = lambda lno: GLOBAL_LINES[lno]
) -> bool:""",
            None,
        ),
        (
            """def potato(
    get_line = ...
):""",
            ["Callable[[int], str]"],
            "bool",
            """def potato(
    get_line: Callable[[int], str] = ...
) -> bool:""",
            None,
        ),
        (
            """def get_timezone_or_utc(
    zone = None
):""",
            ["Union[int, None]"],
            "Union[pytz.tzfile.America/Los_Angeles, int]",
            """def get_timezone_or_utc(
    zone: int | None = None
) -> datetime.tzinfo | int:""",
            None,
        ),
        ("def meep(x: T):", ["T`1"], "T`1", "def meep(x: T) -> T:", None),
        ("def meep(x: T):", ["T`1"], "T-1", "def meep(x: T) -> Any:", None),
        (
            "async def pop_obj(obj_id: int):",
            ["float"],
            "object",
            "async def pop_obj(obj_id: int) -> object:",
            None,
        ),
        (
            "async def pop_obj(self, obj_id: int, /, **kwargs):",
            ["float", "str"],
            "object",
            "async def pop_obj(self, obj_id: int, /, **kwargs: str) -> object:",
            None,
        ),
        (
            "def apply(self, funcname, /, *args, **kwargs):",
            ["str", "Any", "Any"],
            "Any",
            "def apply(self, funcname: str, /, *args: Any, **kwargs: Any) -> Any:",
            None,
        ),
        (
            "def apply(self, function, /, *args, **kwargs):",
            ["Callable[ArgsT, RetT]", "ArgsT.args", "ArgsT.kwargs"],
            "RetT",
            "def apply(self, function: Callable[ArgsT, RetT], /, *args: ArgsT.args, **kwargs: ArgsT.kwargs) -> RetT:",
            None,
        ),
        (
            "def __repr__(self):",
            [],
            "str",
            "def __repr__(self) -> str:",
            None,
        ),
        (
            "def __repr__(self,):",
            [],
            "str",
            "def __repr__(self, ) -> str:",
            None,
        ),
        (
            "def get(valuetype = int):",
            ["type"],
            "object",
            "def get(valuetype: type = int) -> object:",
            None,
        ),
        (
            "def get(valuetype = int,):",
            ["type"],
            "object",
            "def get(valuetype: type = int, ) -> object:",
            None,
        ),
        (
            "def bitfield(self, bits: int = ~ALL_BITS):",
            ["int"],
            "int",
            "def bitfield(self, bits: int = ~ALL_BITS) -> int:",
            None,
        ),
        (
            "def bitfield(self, bits = ~ALL_BITS):",
            ["int"],
            "int",
            "def bitfield(self, bits: int = ~ALL_BITS) -> int:",
            None,
        ),
        (
            "def meep(x: T | None):",
            ["Union[T, None]"],
            "T`1",
            "def meep(x: T | None) -> T:",
            None,
        ),
        (
            "def meep(x):",
            ["Callable[..., Any]"],
            "Any",
            "def meep(x: Callable[..., Any]) -> Any:",
            None,
        ),
        (
            "def read_global(self, length = GLOBALSTART+GLOBALREAD):",
            ["int"],
            "int",
            "def read_global(self, length: int = GLOBALSTART + GLOBALREAD) -> int:",
            None,
        ),
        (
            "def meep(x = A @ B):",
            ["Matrix"],
            "Vector",
            "def meep(x: Matrix = A @ B) -> Vector:",
            None,
        ),
        (
            """def line_len(
    get_line = ...
):""",
            ["Callable[[], str]"],
            "int",
            """def line_len(
    get_line: Callable[[], str] = ...
) -> int:""",
            None,
        ),
        (
            """def line_len(
    get_line = ...
) -> int | None:""",
            ["Callable[[], str]"],
            "Union[int, None]",
            """def line_len(
    get_line: Callable[[], str] = ...
) -> int | None:""",
            None,
        ),
        (
            """def test(value: int = -~ALL_BITS):""",
            ["int"],
            "None",
            """def test(value: int = -~ALL_BITS) -> None:""",
            None,
        ),
        (
            """def potato(
    get_line = lambda lno: "",
    lines = 0,
):""",
            ["Callable[[int], str]", "int"],
            "bool",
            """def potato(
    get_line: Callable[[int], str] = lambda lno: "",
    lines: int = 0,
) -> bool:""",
            None,
        ),
        (
            """def potato(
    get_line = lambda lno, default: (default),
    lines = 0,
):""",
            ["Callable[[int, str], str]", "int"],
            "bool",
            """def potato(
    get_line: Callable[[int, str], str] = lambda lno, default: (default),
    lines: int = 0,
) -> bool:""",
            None,
        ),
        (
            """def potato(
    get_line = lambda lno, default: (default, GLOBAL), lines = 0,
):""",
            ["Callable[[int, str], str]", "int"],
            "bool",
            """def potato(
    get_line: Callable[[int, str], str] = lambda lno, default: (default, GLOBAL), lines: int = 0,
) -> bool:""",
            None,
        ),
        (
            """def potato(
    get_line: Callable[[int, str], str] = lambda lno, default: (default),
    lines: int = 0,
) -> bool:""",
            ["Callable[[int, str], str]", "int"],
            "bool",
            """def potato(
    get_line: Callable[[int, str], str] = lambda lno, default: (default),
    lines: int = 0,
) -> bool:""",
            None,
        ),
        (
            """def potato(get_line: Callable[[int, str], str] = lambda lno, default: default) -> bool:""",
            ["Callable[[int, str], str]"],
            "bool",
            """def potato(get_line: Callable[[int, str], str] = lambda lno, default: default) -> bool:""",
            None,
        ),
        (
            """def wrapper(**kwargs) -> Any:""",
            ["Any"],
            "Any",
            """def wrapper(**kwargs: Any) -> Any:""",
            None,
        ),
        (
            """def frog(numbers = (2, 3)) -> int:""",
            ["Tuple[int, int]"],
            "int",
            """def frog(numbers: tuple[int, int] = (2, 3)) -> int:""",
            None,
        ),
        (
            """def log_active_exception(path = f'{ROOT_DIR}/logs/latest.txt'):""",
            ["str"],
            "None",
            """def log_active_exception(path: str = f'{ROOT_DIR}/logs/latest.txt') -> None:""",
            None,
        ),
        (
            """def log_active_exception(path = f'{ROOT_DIR}/logs/latest.txt'):""",
            ["str"],
            "Overload(int, str)",
            """def log_active_exception(path: str = f'{ROOT_DIR}/logs/latest.txt') -> overload[int, str]:""",
            None,
        ),
        (
            """def valid_moves(turn: bool, lines, boxes) -> Generator[Action, None, None]:""",
            [
                "bool",
                "dots and boxes:Sequence[dots and boxes.Sequence[int]]",
                "dots and boxes:Sequence[dots and boxes.Sequence[int]]",
            ],
            "dots and boxes:Generator[dots and boxes.Action, None, None]",
            "def valid_moves(turn: bool, lines: Sequence[Sequence[int]], boxes: Sequence[Sequence[int]]) -> Generator[Action, None, None]:",
            "dots and boxes",
        ),
        (
            """def bad_default_arg(name_map = {"jerald": "cat", "luigi": "person"}):""",
            [
                "dict[str, str]",
            ],
            "str",
            """def bad_default_arg(name_map: dict[str, str] = {"jerald": "cat", "luigi": "person"}) -> str:""",
            None,
        ),
        (
            """def bad_default_arg(name_map = {"jerald", "cat", "bob"}):""",
            [
                "set[str]",
            ],
            "str",
            """def bad_default_arg(name_map: set[str] = {"jerald", "cat", "bob"}) -> str:""",
            None,
        ),
    ],
)
def test_get_annotation(
    function_text: str,
    arg_types: Sequence[str],
    return_type: str,
    result: str,
    filename: str | None,
) -> None:
    annotation_dict = {
        "line": 0,
        "signature": {"arg_types": arg_types, "return_type": return_type},
    }
    if filename:
        annotation_dict["path"] = f"/{filename}.py"

    lines = function_text.splitlines(True)

    def get_line(line_no: int) -> str:
        if line_no >= len(lines):
            return ""
        return lines[line_no]

    returned, _ = annotate.get_annotation(annotation_dict, get_line)
    if returned != result:
        print(f"{returned}\n!=\n{result}")
    assert returned == result


@pytest.mark.skipif(
    sys.version_info < (3, 12),
    reason="F-string tokenization worked differently prior to 3.12",
)
def test_get_annotation_fstring_3_12() -> None:
    function_text = """def log_active_exception(path = f'{f'{ROOT_DIR}/logs'}/latest.txt'):"""
    arg_types = ["str"]
    return_type = "None"
    result = """def log_active_exception(path: str = f'{f'{ROOT_DIR}/logs'}/latest.txt') -> None:"""
    annotation_dict = {
        "line": 0,
        "signature": {"arg_types": arg_types, "return_type": return_type},
    }

    lines = function_text.splitlines(True)

    def get_line(line_no: int) -> str:
        return lines[line_no]

    returned, _ = annotate.get_annotation(annotation_dict, get_line)
    if returned != result:
        print(f"{returned}\n!=\n{result}")
    assert returned == result


# I don't think it's possible for f-string to fail without tokenization errors
# before we get to that point.
# @pytest.mark.skipif(
#     sys.version_info < (3, 12),
#     reason="F-string tokenization worked differently prior to 3.12",
# )
# def test_read_fstring_failure() -> None:
#     text = '''f'Happy birthday, {username}!{f"}'; pass'''
#
#     with StringIO(text) as file:
#         tokenize_gen = generate_tokens(file.readline)
#         first_read = next(tokenize_gen)
#         assert first_read.type == 61 # (FSTRING_START)
#         assert first_read.string == "f'"
#         with pytest.raises(
#             annotate.ParseError,
#             match="Reading f-string failed",
#         ):
#             print(annotate.read_fstring(first_read, tokenize_gen))


def test_get_annotation_tokenization_falure() -> None:
    annotation_dict = {
        "line": 0,
    }

    lines = "def def".splitlines(True)

    def get_line(line_no: int) -> str:
        return lines[line_no]

    with pytest.raises(
        annotate.ParseError,
        match="Did not expect second definition keyword",
    ):
        annotate.get_annotation(annotation_dict, get_line)


def test_read_lambda_failure() -> None:
    text = """ x: GLOBAL_MAP[x]"""

    with StringIO(text) as file:
        with pytest.raises(
            annotate.ParseError,
            match="Reading lambda failed",
        ):
            annotate.read_lambda(generate_tokens(file.readline))


def test_read_lambda_over_read_detection() -> None:
    text = """ line_no, default: f'{line_no}: {default}') -> None:"""

    with StringIO(text) as file:
        content, over_read_error = annotate.read_lambda(
            generate_tokens(file.readline),
        )
        assert content == " line_no, default: f'{line_no}: {default}'"
        assert over_read_error


def test_read_lambda_over_read_parens() -> None:
    text = """ line_no, default: f'{line_no}: {default}'):"""

    with StringIO(text) as file:
        content, over_read_error = annotate.read_lambda(
            generate_tokens(file.readline),
        )
        assert content == " line_no, default: f'{line_no}: {default}'"
        assert over_read_error


@pytest.mark.parametrize(
    "end",
    [
        "\n",
        ",",
    ],
)
def test_read_lambda(end: str) -> None:
    text = f""" x: GLOBAL_MAP[x]{end}"""

    with StringIO(text) as file:
        content, over_read_error = annotate.read_lambda(
            generate_tokens(file.readline),
        )
        assert not over_read_error
        assert content == text[:-1], f"{content!r} != {text!r}"


def test_get_annotation_tokenization_eof_falure() -> None:
    annotation_dict = {
        "line": 0,
    }

    lines = "def waffle() -> str".splitlines(True)

    def get_line(line_no: int) -> str:
        if line_no > 0:
            return ""
        return lines[line_no]

    with pytest.raises(
        annotate.ParseError,
        match="Reached End of File, expected end of definition",
    ):
        annotate.get_annotation(annotation_dict, get_line)


def test_parse_type_list_no_close_failure() -> None:
    parser = annotate.Parser(
        [annotate.DottedName("int"), annotate.Separator("notend")],
    )
    with pytest.raises(
        annotate.ParseError,
        match="Expected '\\)' or '\\]', got 'notend'",
    ):
        parser.parse_type_list()


def test_parse_union_list() -> None:
    parser = annotate.Parser(
        [
            annotate.DottedName("int"),
            annotate.Separator("|"),
            annotate.DottedName("str"),
            annotate.End(),
        ],
    )
    parser.parse_union_list()


def test_parse_single_no_empty_union_failure() -> None:
    parser = annotate.Parser(
        [
            annotate.DottedName("Union"),
            annotate.Separator("["),
            annotate.Separator("]"),
        ],
    )
    with pytest.raises(annotate.ParseError, match="No items in Union"):
        parser.parse_single()


def test_parse_single_no_null_collection_failure() -> None:
    parser = annotate.Parser(
        [
            annotate.DottedName(),
        ],
    )
    with pytest.raises(
        annotate.ParseError,
        match="Expected token text to be string, got None",
    ):
        parser.parse_single()


def test_peek_out_of_tokens_failure() -> None:
    parser = annotate.Parser([])
    with pytest.raises(
        annotate.ParseError,
        match="Ran out of tokens",
    ):
        parser.peek()


def test_expect_failure() -> None:
    parser = annotate.Parser([annotate.DottedName("waffle")])
    with pytest.raises(
        annotate.ParseError,
        match="Expected 'def', got 'waffle'",
    ):
        parser.expect("def")


def test_expect_type_failure() -> None:
    parser = annotate.Parser([annotate.DottedName("waffle")])
    with pytest.raises(
        annotate.ParseError,
        match="Expected 'ReturnTypeDef', got DottedName\\(text='waffle'\\)",
    ):
        parser.expect_type(annotate.ReturnTypeDef)


def test_expect_type_multi_failure() -> None:
    parser = annotate.Parser([annotate.DottedName("waffle")])
    with pytest.raises(
        annotate.ParseError,
        match="Expected 'ReturnTypeDef' or 'Separator', got DottedName\\(text='waffle'\\)",
    ):
        parser.expect_type((annotate.ReturnTypeDef, annotate.Separator))


@pytest.mark.parametrize(
    ("tokens", "value"),
    [
        (
            [
                annotate.DottedName("mypy_extensions.NoReturn"),
            ],
            annotate.TypeValue("NoReturn"),
        ),
        (
            [
                annotate.DottedName("typing.NoReturn"),
            ],
            annotate.TypeValue("NoReturn"),
        ),
        (
            [
                annotate.DottedName("Tuple"),
                annotate.Separator("["),
                annotate.DottedName("int"),
                annotate.Separator(","),
                annotate.DottedName("str"),
                annotate.Separator("]"),
            ],
            annotate.TypeValue(
                "Tuple",
                (
                    annotate.TypeValue("int"),
                    annotate.TypeValue("str"),
                ),
            ),
        ),
        (
            [
                annotate.DottedName("Union"),
                annotate.Separator("["),
                annotate.DottedName("int"),
                annotate.Separator("]"),
            ],
            annotate.TypeValue("int"),
        ),
        (
            [
                annotate.DottedName("Optional"),
                annotate.Separator("["),
                annotate.DottedName("int"),
                annotate.Separator("]"),
            ],
            annotate.TypeValue(
                "Union",
                (annotate.TypeValue("int"), annotate.TypeValue("None")),
            ),
        ),
        (
            [
                annotate.DottedName("Callable"),
                annotate.Separator("["),
                annotate.Separator("["),
                annotate.Separator("]"),
                annotate.Separator(","),
                annotate.DottedName("int"),
                annotate.Separator("]"),
            ],
            annotate.TypeValue(
                "Callable",
                (annotate.TypeValue("[]"), annotate.TypeValue("int")),
            ),
        ),
    ],
)
def test_parse_single(
    tokens: list[annotate.Token],
    value: annotate.TypeValue,
) -> None:
    parser = annotate.Parser(tokens)

    result = parser.parse_single()

    if value != result:
        print(f"{value}\n!=\n{result}")
    assert result == result



"""Thread-safe version of Tkinter.

Copyright (c) 2009, Allen B. Taylor
Copyright (c) 2017, baldk
Copyright (c) 2018, RedFantom

This module is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser Public License for more details.

You should have received a copy of the GNU Lesser Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Usage:

    import mtTkinter as Tkinter
    # Use "Tkinter." as usual.

or

    from mtTkinter import *
    # Use Tkinter module definitions as usual.

This module modifies the original Tkinter module in memory, making all
functionality thread-safe. It does this by wrapping the Tk class' tk
instance with an object that diverts calls through an event queue when
the call is issued from a thread other than the thread in which the Tk
instance was created. The events are processed in the creation thread
via an 'after' event.

Note that, because it modifies the original Tkinter module (in memory),
other modules that use Tkinter (e.g., Pmw) reap the benefits automagically
as long as mtTkinter is imported at some point before extra threads are
created.

Authors:
Allen B. Taylor, a.b.taylor@gmail.com
RedFantom, redfantom@outlook.com
baldk, baldk@users.noreply.github.com

Docstrings and line-comments wrapped to 80 characters, code wrapped to
100 characters.
"""

from __future__ import annotations

import queue
import threading
from tkinter import TRUE as TRUE, Tk
from typing import TYPE_CHECKING, Any, cast

if TYPE_CHECKING:
    from collections.abc import Callable
    from types import TracebackType

# Use TRUE somewhere so pycln doesn't eat it
assert TRUE


class _Tk:
    """Wrapper for underlying attribute tk of class Tk."""

    def __init__(
        self,
        tk: Tk,
        mt_debug: int = 0,
        mt_check_period: int = 10,
    ) -> None:
        """Initialize Tk Wrapper.

        :param tk: Tkinter.Tk.tk Tk interpreter object
        :param mt_debug: Determines amount of debug output.
            0 = No debug output (default)
            1 = Minimal debug output
            ...
            9 = Full debug output
        :param mt_check_period: Amount of time in milliseconds (default
            10) between checks for out-of-thread events when things are
            otherwise idle. Decreasing this value can improve GUI
            responsiveness, but at the expense of consuming more CPU
            cycles.

        # TODO: Replace custom logging functionality with standard
        # TODO: logging.Logger for easier access and standardization
        """
        self._tk = tk

        # Create the incoming event queue
        self._event_queue: queue.Queue[
            tuple[
                Any,
                tuple[Any, ...],
                dict[str, Any],
                queue.Queue[
                    tuple[
                        bool,
                        tuple[
                            type[BaseException],
                            BaseException,
                            TracebackType,
                        ]
                        | Any,
                    ]
                ],
            ]
        ] = queue.Queue(1)

        # Identify the thread from which this object is being created
        # so we can tell later whether an event is coming from another
        # thread.
        self._creation_thread = threading.current_thread()

        # Create attributes for kwargs
        self._debug = mt_debug
        self._check_period = mt_check_period
        # Destroying flag to be set by the .destroy() hook
        self._destroying = False

    def __getattr__(self, name: str) -> _TkAttr | Any:
        """Diverts attribute accesses to a wrapper around the underlying tk object."""
        ##        if name in {"createcommand", "call"}:
        attr = getattr(self._tk, name)
        if callable(attr):
            return _TkAttr(self, getattr(self._tk, name))
        return attr
        ##        return getattr(self._tk, name)


class _TkAttr:
    """Thread-safe callable attribute wrapper."""

    def __init__(self, tk: _Tk, attr: Callable[..., Any]) -> None:
        self._tk = tk
        self._attr = attr

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Thread-safe method invocation.

        Diverts out-of-thread calls through the event queue.
        Forwards all other method calls to the underlying tk object directly.
        """
        # Check if we're in the creation thread
        if threading.current_thread() == self._tk._creation_thread:
            # We're in the creation thread; just call the event directly
            if (
                self._tk._debug >= 8
                or self._tk._debug >= 3
                and self._attr.__name__ == "call"
                and len(args) >= 1
                and args[0] == "after"
            ):
                print(
                    "Calling event directly:",
                    self._attr.__name__,
                    args,
                    kwargs,
                )
            return self._attr(*args, **kwargs)
        if not self._tk._destroying:
            # We're in a different thread than the creation thread;
            # enqueue the event, and then wait for the response.
            response_queue: queue.Queue[
                tuple[
                    bool,
                    tuple[type[BaseException], BaseException, TracebackType]
                    | Any,
                ]
            ] = queue.Queue(1)
            if self._tk._debug >= 1:
                print(
                    "Marshalling event:",
                    self._attr.__name__,
                    args,
                    kwargs,
                )
            self._tk._event_queue.put(
                (self._attr, args, kwargs, response_queue),
                True,
                1,
            )
            is_exception, response = response_queue.get(True, None)

            # Handle the response, whether it's a normal return value or
            # an exception.
            if is_exception:
                ex_type, ex_value, ex_tb = response
                raise ex_type(ex_value, ex_tb)
            return response
        return None


def _tk__init__(self: Tk, *args: Any, **kwargs: Any) -> None:
    """Overwrite Tkinter.Tk.__init__ method.

    :param self: Tk instance
    :param args, kwargs: Arguments for Tk initializer.
    """
    # We support some new keyword arguments that the original __init__ method
    # doesn't expect, so separate those out before doing anything else.
    new_kwnames = ("mt_check_period", "mt_debug")
    new_kwargs = {
        kw_name: kwargs.pop(kw_name)
        for kw_name in new_kwnames
        if kwargs.get(kw_name, None) is not None
    }

    # Call the original __init__ method, creating the internal tk member.
    assert hasattr(self, "__original__init__mtTkinter")
    self.__original__init__mtTkinter(*args, **kwargs)

    # Replace the internal tk member with a wrapper that handles calls from
    # other threads.
    self.tk = _Tk(self.tk, **new_kwargs)  # type: ignore[assignment,arg-type]

    # Set up the first event to check for out-of-thread events.
    self.after_idle(_check_events, self)


# Define a hook for class Tk's destroy method.
def _tk_destroy(self: Tk) -> None:
    if hasattr(self.tk, "_destroying"):
        self.tk._destroying = True
    if hasattr(self, "__original__destroy"):
        self.__original__destroy()


def _check_events(tk: Tk) -> None:
    """Check events in the queue on a given Tk instance."""
    used = False
    tk_obj = cast(_Tk, tk.tk)
    try:
        # Process all enqueued events, then exit.
        while True:
            try:
                # Get an event request from the queue.
                method, args, kwargs, response_queue = (
                    tk_obj._event_queue.get_nowait()
                )
            except queue.Empty:
                # No more events to process.
                break
            else:
                # Call the event with the given arguments, and then return
                # the result back to the caller via the response queue.
                used = True
                if tk_obj._debug >= 2:
                    print(
                        "Calling event from main thread:",
                        method.__name__,
                        args,
                        kwargs,
                    )
                try:
                    response_queue.put((False, method(*args, **kwargs)))
                except SystemExit:
                    raise  # Raises original SystemExit
                except Exception:
                    # Calling the event caused an exception; return the
                    # exception back to the caller so that it can be raised
                    # in the caller's thread.
                    from sys import exc_info  # Python 2 requirement

                    ex_type, ex_value, ex_tb = exc_info()
                    response_queue.put((True, (ex_type, ex_value, ex_tb)))
    finally:
        # Schedule to check again. If we just processed an event, check
        # immediately; if we didn't, check later.
        if used:
            tk.after_idle(_check_events, tk)
        else:
            tk.after(tk_obj._check_period, _check_events, tk)


def restore() -> None:
    """Restore original functions."""
    if hasattr(Tk, "__original__init__mtTkinter"):
        Tk.__init__ = Tk.__original__init__mtTkinter  # type: ignore[method-assign]
        del Tk.__original__init__mtTkinter

    if hasattr(Tk, "__original__destroy"):
        Tk.destroy = Tk.__original__destroy  # type: ignore[method-assign]
        del Tk.__original__destroy


"""Perform in-memory modification of Tkinter module"""
# Replace Tk's original __init__ with the hook.
if not hasattr(Tk, "__original__init__mtTkinter"):
    Tk.__original__init__mtTkinter = Tk.__init__  # type: ignore[attr-defined]
    Tk.__init__ = _tk__init__  # type: ignore[method-assign]

# Replace Tk's original destroy with the hook.
if not hasattr(Tk, "__original__destroy"):
    Tk.__original__destroy = Tk.destroy  # type: ignore[attr-defined]
    Tk.destroy = _tk_destroy  # type: ignore[method-assign]



from __future__ import annotations

import pytest

from idlemypyextension import utils


def test_get_required_config() -> None:
    assert (
        utils.get_required_config(
            {
                "fish_enabled": "True",
                "waffle_mode": "False",
                "user_settings": "1, 2, 3",
            },
            {
                "check_fishes": "Ctrl+Alt+f",
                "reboot": "Ctrl+Alt+Del",
            },
            "fish_extend",
        )
        == """
[fish_extend]
fish_enabled = True
waffle_mode = False
user_settings = 1, 2, 3

[fish_extend_cfgBindings]
check_fishes = Ctrl+Alt+f
reboot = Ctrl+Alt+Del"""
    )


def test_get_line_selection() -> None:
    assert utils.get_line_selection(3) == ("3.0", "4.0")
    assert utils.get_line_selection(3, 3) == ("3.0", "6.0")


@pytest.mark.parametrize(
    ("line", "expected"),
    [
        ("0.0", (0, 0)),
        ("1.0", (1, 0)),
        ("0.1", (0, 1)),
        ("29120.314817317", (29120, 314817317)),
        ("-231.-234327", (-231, -234327)),
    ],
)
def test_get_line_col_success(line: str, expected: tuple[int, int]) -> None:
    assert utils.get_line_col(line) == expected


@pytest.mark.parametrize(
    "line",
    [
        "",
        "93732",
        "0x7f",
        "0x74.0xff",
    ],
)
def test_get_line_col_failure(line: str) -> None:
    with pytest.raises(
        ValueError,
        match=r"(not enough values to unpack)|(invalid literal for int\(\) with base 10:)",
    ):
        utils.get_line_col(line)


@pytest.mark.parametrize(
    ("index", "offset", "expect"),
    [("3.14", 0, "3.0"), ("3.14", 1, "4.0"), ("2981.23", -1, "2980.0")],
)
def test_get_whole_line(index: str, offset: int, expect: str) -> None:
    assert utils.get_whole_line(index, offset) == expect


@pytest.mark.parametrize(
    ("text", "expect"),
    [("  waf", 2), ("cat", 0), ("     fish", 5), ("   ", 3), ("", 0)],
)
def test_get_line_indent(text: str, expect: int) -> None:
    assert utils.get_line_indent(text, " ") == expect



"""Type Check IDLE Extension."""

# Programmed by CoolCat467

from __future__ import annotations

# IDLE Mypy daemon integration extension
# Copyright (C) 2023-2024  CoolCat467
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__title__ = "extension"
__author__ = "CoolCat467"
__license__ = "GNU General Public License Version 3"

import json
import math
import os
import re
import sys
import traceback
from functools import partial, wraps
from idlelib.config import idleConf
from typing import TYPE_CHECKING, Any, ClassVar, Final

from idlemypyextension import annotate, client, mttkinter, tktrio, utils

if TYPE_CHECKING:
    from collections.abc import Callable
    from idlelib.pyshell import PyShellEditorWindow
    from tkinter import Event

DAEMON_TIMEOUT_MIN: Final = 5
ACTION_TIMEOUT_MIN: Final = 5


def debug(message: object) -> None:
    """Print debug message."""
    # TODO: Censor username/user files
    print(f"\n[{__title__}] DEBUG: {message}")


def parse_comments(
    mypy_output: str,
    default_file: str,
    default_line: int,
) -> dict[str, list[utils.Comment]]:
    """Parse mypy output, return mapping of filenames to lists of comments."""
    error_type = re.compile(r"  \[[a-z\-]+\]\s*$")

    files: dict[str, list[utils.Comment]] = {}
    for output_line in mypy_output.splitlines():
        if not output_line.strip():
            continue
        filename = default_file
        line = default_line
        line_end = default_line
        col = 0
        col_end = 0
        msg_type = "unrecognized"

        if output_line.count(": ") < 2:
            text = output_line
        else:
            where, msg_type, text = output_line.split(": ", 2)

            position = where.split(":")

            filename = position[0]
            if len(position) > 1:
                line = int(position[1])
                line_end = line
            if len(position) > 2:
                col = int(position[2])
                col_end = col
            if len(position) > 4:
                line_end = int(position[3])
                if line_end == line:
                    col_end = int(position[4])
                else:
                    line_end = line
        comment_type = error_type.search(text)
        if comment_type is not None:
            text = text[: comment_type.start()]
            msg_type = f"{comment_type.group(0)[3:-1]} {msg_type}"

        comment = utils.Comment(
            file=filename,
            line=line,
            contents=f"{msg_type}: {text}",
            column=col,
            line_end=line_end,
            column_end=col_end,
        )

        files.setdefault(filename, [])
        files[filename].append(comment)
    return files


# Important weird: If event handler function returns 'break',
# then it prevents other bindings of same event type from running.
# If returns None, normal and others are also run.


class idlemypyextension(utils.BaseExtension):  # noqa: N801
    """Add comments from mypy to an open program."""

    __slots__ = ("triorun",)
    # Extend the file and format menus.
    menudefs: ClassVar = [
        (
            "edit",
            [
                None,
                ("_Type Check File", "<<type-check>>"),
                ("Find Next Type Comment", "<<find-next-type-comment>>"),
            ],
        ),
        (
            "format",
            [
                ("Suggest Signature", "<<suggest-signature>>"),
                ("Remove Type Comments", "<<remove-type-comments>>"),
            ],
        ),
        ("run", [("Shutdown dmypy daemon", "<<shutdown-dmypy-daemon>>")]),
    ]
    # Default values for configuration file
    values: ClassVar = {
        "enable": "True",
        "enable_editor": "True",
        "enable_shell": "False",
        "daemon_flags": "None",
        "search_wrap": "True",
        "suggest_replace": "False",
        "timeout_mins": "30",
        "action_max_sec": "None",
    }
    # Default key binds for configuration file
    bind_defaults: ClassVar = {
        "type-check": "<Alt-Key-t>",
        "suggest-signature": "<Alt-Key-s>",
        "remove-type-comments": "<Alt-Shift-Key-T>",
        "find-next-type-comment": "<Alt-Key-g>",
    }

    # Overwritten in reload
    daemon_flags = "None"
    search_wrap = "True"
    suggest_replace = "False"
    timeout_mins = "30"
    action_max_sec = "None"

    # Class attributes
    idlerc_folder = os.path.expanduser(idleConf.userdir)
    mypy_folder = os.path.join(idlerc_folder, "mypy")
    status_file = os.path.join(mypy_folder, "dmypy.json")
    log_file = os.path.join(mypy_folder, "log.txt")

    def __init__(self, editwin: PyShellEditorWindow) -> None:
        """Initialize the settings for this extension."""
        super().__init__(editwin, comment_prefix="types")

        if not os.path.exists(self.mypy_folder):
            os.mkdir(self.mypy_folder)

        self.triorun = tktrio.TkTrioRunner(
            self.editwin.top,
            self.editwin.close,
        )

        for attr_name in dir(self):
            if attr_name.startswith("_"):
                continue
            if attr_name.endswith("_event_async"):
                bind_name = "-".join(attr_name.split("_")[:-2]).lower()
                self.text.bind(f"<<{bind_name}>>", self.get_async(attr_name))
                # print(f'{attr_name} -> {bind_name}')

    def get_async(
        self,
        name: str,
    ) -> Callable[[Event[Any]], str]:
        """Get sync callable to run async function."""
        async_function = getattr(self, name)

        @wraps(async_function)
        @utils.log_exceptions
        def call_trio(event: Event[Any]) -> str:
            self.triorun(partial(async_function, event))
            return "break"

        return call_trio

    @property
    def daemon_timeout(self) -> int:
        """Daemon timeout."""
        if self.timeout_mins == "None":
            return DAEMON_TIMEOUT_MIN * 60
        try:
            return max(
                DAEMON_TIMEOUT_MIN * 60,
                math.ceil(float(self.timeout_mins) * 60),
            )
        except ValueError:
            return DAEMON_TIMEOUT_MIN * 60

    @property
    def action_timeout(self) -> int | None:
        """Action timeout."""
        if self.action_max_sec == "None":
            return None
        try:
            return max(ACTION_TIMEOUT_MIN, int(self.action_max_sec))
        except ValueError:
            return max(ACTION_TIMEOUT_MIN, int(self.values["action_max_sec"]))

    @property
    def flags(self) -> list[str]:
        """Mypy Daemon flags."""
        base = {
            "--hide-error-context",
            "--no-color-output",
            "--show-absolute-path",
            "--no-error-summary",
            "--soft-error-limit=-1",
            "--show-traceback",
            f'--cache-dir="{self.mypy_folder}"',
            # f'--log-file="{self.log_file}"',
            # "--cache-fine-grained",
        }
        if self.daemon_flags == "None":
            return list(base)
        extra = set()
        for arg in self.daemon_flags.split(" "):
            value = arg.strip()
            if value:
                extra.add(value)
        return list(base | extra)

    @property
    def typecomment_only_current_file(self) -> bool:
        """Should only add type comments for currently open file?."""
        return True

    def add_type_comments_for_file(
        self,
        target_filename: str,
        comments: list[utils.Comment],
    ) -> dict[str, list[int]]:
        """Add type comments for target files.

        Return list of lines were a comment was added.
        """
        # Split up comments by line in order
        line_data: dict[int, list[utils.Comment]] = {}
        for comment in comments:
            line_data.setdefault(comment.line, [])
            line_data[comment.line].append(comment)

        all_messages = []
        for line in sorted(line_data):
            messages = line_data[line]
            if not messages:
                continue
            all_messages.extend(messages)
            pointers = self.get_pointers(messages)
            if pointers is not None:
                all_messages.append(pointers)

        return self.add_comments(all_messages)

    def add_mypy_messages(
        self,
        start_line: int,
        mypy_output: str,
        only_filename: str | None = None,
    ) -> dict[str, list[int]]:
        """Add mypy comments for target filename.

        Return list of lines where comments were added.
        """
        assert self.files.filename is not None

        if only_filename is not None:
            only_filename = os.path.abspath(only_filename)

        files = parse_comments(
            mypy_output,
            os.path.abspath(self.files.filename),
            start_line,
        )

        file_commented_lines: dict[str, list[int]] = {}

        to_comment = list(files)

        if self.typecomment_only_current_file:
            assert only_filename is not None
            to_comment = [only_filename]

            # Find first line in target file or use start_line
            if not files.get(only_filename):
                other_files_comment_line = start_line
            else:
                other_files_comment_line = min(
                    comment.line for comment in files[only_filename]
                )

            # Add comments about how other files have errors
            files.setdefault(only_filename, [])
            for filename in files:
                if filename == only_filename:
                    continue
                files[only_filename].append(
                    utils.Comment(
                        file=only_filename,
                        line=other_files_comment_line,
                        contents=f"Another file has errors: {filename!r}",
                        column_end=0,
                    ),
                )

        for target_filename in to_comment:
            if target_filename not in files:
                continue
            file_comments = self.add_type_comments_for_file(
                target_filename,
                files[target_filename],
            )
            file_commented_lines.update(file_comments)
        return file_commented_lines

    def add_extra_data(
        self,
        file: str,
        start_line: int,
        data: str,
        prefix: str = "",
    ) -> tuple[int, list[int]]:
        """Add extra data to file as a big block of comments.

        Returns
        -------
        Tuple of:
        - Number of lines attempted to add
        - List of line numbers added that were not already there
        otherwise empty because no content.

        """
        if not data:
            return 0, []
        lines = data.splitlines()
        if not lines:
            return 0, []
        lines[0] = f"{prefix}{lines[0]}"
        added = self.add_comment_block(file, start_line, lines)
        return len(lines), added

    def add_errors(
        self,
        file: str,
        start_line: int,
        errors: str,
    ) -> tuple[int, list[int]]:
        """Add error lines to file as a block of comments.

        Returns
        -------
        Tuple of:
        - Number of lines attempted to add
        - List of line numbers added that were not already there
        otherwise None because no content.

        """
        return self.add_extra_data(
            file,
            start_line,
            errors,
            prefix="Error running mypy: ",
        )

    async def ensure_daemon_running(self) -> bool:
        """Make sure daemon is running. Return False if cannot continue."""
        if not client.is_running(self.status_file):
            return await client.start(
                self.status_file,
                flags=self.flags,
                daemon_timeout=self.daemon_timeout,
                log_file=self.log_file,
            )
        return True

    async def shutdown_dmypy_daemon_event_async(
        self,
        event: Event[Any],
    ) -> str:
        """Shutdown dmypy daemon event handler."""
        # pylint: disable=unused-argument
        if not client.is_running(self.status_file):
            self.text.bell()
            return "break"

        # Only stop if running
        response = await client.stop(self.status_file)
        if response.get("err") or response.get("error"):
            # Kill
            client.kill(self.status_file)

        return "break"

    async def check(self, file: str) -> client.Response:
        """Perform dmypy check."""
        if not await self.ensure_daemon_running():
            return client.Response(
                {"out": "", "err": "Error: Could not start mypy daemon"},
            )
        flags = self.flags
        flags += [file]
        # debug(f"check {flags = }")
        command = " ".join(
            x
            for x in [
                "dmypy",
                f'--status-file="{self.status_file}"',
                "run",
                (
                    f"--timeout={self.action_timeout}"
                    if self.action_timeout
                    else ""
                ),
                f'--log-file="{self.log_file}"',
                "--export-types",
                f'"{file}"',
                "--",
                *self.flags,
            ]
            if x
        )
        debug(f"{command = }")
        return await client.run(
            self.status_file,
            flags=flags,
            timeout=self.action_timeout,
            daemon_timeout=self.daemon_timeout,
            log_file=self.log_file,
            export_types=True,
        )

    def get_suggestion_text(
        self,
        annotation: dict[str, object],
    ) -> tuple[str | None, int]:
        """Get suggestion text from annotation.

        Return None on error or no difference, text if different
        """
        # while annotation["line"] >= 0 and "def" not in self.get_line(
        #     annotation["line"],
        # ):
        #     annotation["line"] -= 1
        line = annotation["line"]
        assert isinstance(line, int)

        try:
            text, line_count = annotate.get_annotation(
                annotation,
                self.get_line,
            )
        except Exception as ex:
            ex_text, ex_traceback = sys.exc_info()[1:]
            traceback.print_exception(
                None,  # Ignored since python 3.5
                value=ex_text,
                tb=ex_traceback,
                limit=None,
                chain=True,
            )
            indent = utils.get_line_indent(self.get_line(line))
            return (
                self.get_comment_line(
                    indent,
                    f"Error generating suggestion: {ex}",
                ),
                1,
            )

        select_start = f"{line}.0"
        line_end = line + line_count
        select_end = f"{line_end}.0"

        if text == self.text.get(select_start, select_end)[:-1]:
            return None, line_count
        return text, line_count

    async def suggest(self, file: str, line: int) -> None:
        """Perform dmypy suggest."""
        if not await self.ensure_daemon_running():
            response = client.Response(
                {"err": "Error: Could not start mypy daemon"},
            )
        else:
            function = f"{file}:{line}"

            command = " ".join(
                (
                    "dmypy",
                    f'--status-file="{self.status_file}"',
                    "suggest",
                    f'"{function}"',
                ),
            )
            debug(f"{command = }")

            response = await client.suggest(
                self.status_file,
                function=function,
                do_json=True,
                timeout=self.action_timeout,
            )
        debug(f"suggest {response = }")

        errors = ""
        if response.get("error"):
            errors += response["error"]
        if response.get("err"):
            if errors:
                errors += "\n\n"
            errors += response["err"]
        if response.get("stderr"):
            if errors:
                errors += "\n\n"
            errors += f'stderr:\n{response["stderr"]}'
        if "out" not in response and not errors:
            errors += "No response from dmypy daemon."

        # Display errors
        if errors:
            # Add mypy errors
            self.add_errors(file, self.editwin.getlineno(), errors)

            self.text.bell()
            return

        annotations = json.loads(response["out"])

        line = annotations[0]["line"]

        samples: dict[int, list[str]] = {}
        line_count = 0
        for annotation in annotations:
            count = annotation["samples"]
            text, suggest_lines = self.get_suggestion_text(annotation)
            if text is None:
                continue
            if count not in samples:
                samples[count] = []
            samples[count].append(text)
            line_count += suggest_lines

        order = sorted(samples, reverse=True)
        lines = []
        for count in order:
            for sample in samples[count]:
                if sample not in lines:
                    lines.append(sample)

        replace = self.suggest_replace == "True"

        if len(lines) == 1:
            text = lines[0]
            if "Error generating suggestion: " in text:
                replace = False
        else:
            text = "\n".join(lines)
            replace = False

        select_start = f"{line}.0"
        line_end = line + line_count
        select_end = f"{line_end}.0"

        if not text or text == self.text.get(select_start, select_end)[:-1]:
            # Bell to let user know happened, just nothing to do
            self.editwin.gotoline(line)
            self.text.bell()
            return

        if not replace and "Error generating suggestion: " not in text:
            text = "\n".join(f"##{line}" for line in text.splitlines())
        text += "\n"

        with utils.undo_block(self.undo):
            if replace:
                self.text.delete(select_start, select_end)

            self.text.insert(select_start, text, ())

        self.editwin.gotoline(line)
        self.text.bell()

    def initial(self) -> tuple[str | None, str | None]:
        """Do common initial setup. Return error or none, file.

        Reload configuration, make sure file is saved,
        and make sure mypy is installed
        """
        # Reload configuration
        self.reload()

        # Get file we are checking
        raw_filename: str | None = self.files.filename
        if raw_filename is None:
            return "break", None
        file: str = os.path.abspath(raw_filename)

        # Make sure file is saved.
        if not self.files.get_saved():
            if not utils.ask_save_dialog(self.text):
                # If not ok to save, do not run. Would break file.
                self.text.bell()
                return "break", file
            # Otherwise, we are clear to save
            self.files.save(None)
            if not self.files.get_saved():
                return "break", file

        # Everything worked
        return None, file

    async def suggest_signature_event_async(self, event: Event[Any]) -> str:
        """Handle suggest signature event."""
        # pylint: disable=unused-argument
        init_return, file = self.initial()

        if init_return is not None:
            return init_return
        if file is None:
            return "break"
        if client.REQUEST_LOCK.locked():
            # If already requesting something from daemon,
            # do not send any other requests.
            debug(client.REQUEST_LOCK.statistics())
            # Make bell sound so user knows this ran even though
            # nothing happened.
            self.text.bell()
            return "break"

        await self.suggest(file, self.editwin.getlineno())

        return "break"

    def type_check_add_response_comments(
        self,
        response: client.Response,
        file: str,
    ) -> None:
        """Add all the comments (error and regular) from dmypy response."""
        debug(f"type check {response = }")

        if "out" in response:
            # Add code comments
            self.add_mypy_messages(
                self.editwin.getlineno(),
                response["out"],
                file,
            )
        if "error" in response:
            self.add_errors(file, self.editwin.getlineno(), response["error"])
        if "err" in response:
            # Add mypy run errors
            self.add_errors(file, self.editwin.getlineno(), response["err"])
        if "stdout" in response:
            self.add_extra_data(
                file,
                self.editwin.getlineno(),
                response["stdout"],
                prefix="dmypy run stdout: ",
            )
        if "stderr" in response:
            self.add_extra_data(
                file,
                self.editwin.getlineno(),
                response["stderr"],
                prefix="dmypy run stderr: ",
            )

        # Make bell sound so user knows we are done,
        # as it freezes a bit while mypy looks at the file
        self.text.bell()

    async def type_check_event_async(self, event: Event[Any]) -> str:
        """Perform a mypy check and add comments."""
        init_return, file = self.initial()

        if init_return is not None:
            return init_return

        if file is None:
            return "break"

        if client.REQUEST_LOCK.locked():
            # If already requesting something from daemon,
            # do not send any other requests.
            debug(client.REQUEST_LOCK.statistics())
            # Make bell sound so user knows this ran even though
            # nothing happened.
            self.text.bell()
            return "break"

        # Run mypy on open file
        response = await self.check(file)

        self.type_check_add_response_comments(response, file)
        return "break"

    @utils.log_exceptions
    def remove_type_comments_event(self, _event: Event[Any]) -> str:
        """Remove selected extension comments."""
        self.remove_selected_extension_comments()
        return "break"

    @utils.log_exceptions
    def remove_all_type_comments(self, _event: Event[Any]) -> str:
        """Remove all extension comments."""
        self.remove_all_extension_comments()
        return "break"

    @utils.log_exceptions
    def find_next_type_comment_event(self, _event: Event[Any]) -> str:
        """Find next extension comment by hacking the search dialog engine."""
        # Reload configuration
        self.reload()

        # Find comment
        self.find_next_extension_comment(self.search_wrap == "True")

        return "break"

    def unregister_async_events(self) -> None:
        """Unregister asynchronous event handlers."""
        for attr_name in dir(self):
            if attr_name.startswith("_"):
                continue
            if attr_name.endswith("_event_async"):
                bind_name = "-".join(attr_name.split("_")[:-2]).lower()
                self.text.event_delete(f"<<{bind_name}>>")

    @utils.log_exceptions
    def close(self) -> None:
        """Extension cleanup before IDLE window closes."""
        # Wrapped in try except so failure doesn't cause zombie windows.
        del self.triorun
        mttkinter.restore()
        try:
            self.unregister_async_events()
        except Exception as exc:
            traceback.print_exception(exc)



"""TKTrio - Run trio on top of Tkinter."""

# Programmed by CoolCat467

from __future__ import annotations

# TKTrio - Run Trio on top of Tkinter.
# Copyright (C) 2023  CoolCat467
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__title__ = "TKTrio"
__author__ = "CoolCat467"
__license__ = "GNU General Public License Version 3"
__version__ = "0.1.0"

import contextlib
import sys
import tkinter as tk
import weakref
from enum import IntEnum, auto
from functools import partial, wraps
from tkinter import messagebox
from traceback import format_exception, print_exception
from typing import TYPE_CHECKING, Any, TypeGuard

from idlemypyextension import mttkinter, utils
from idlemypyextension.moduleguard import guard_imports

with guard_imports({"trio", "exceptiongroup"}):
    import trio

    if sys.version_info < (3, 11):
        from exceptiongroup import ExceptionGroup

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from outcome import Outcome
    from typing_extensions import Self


# Use mttkinter somewhere so pycln doesn't eat it
assert mttkinter.TRUE


def debug(message: object) -> None:
    """Print debug message."""
    # TODO: Censor username/user files
    print(f"\n[{__title__}] DEBUG: {message}")


def install_protocol_override(
    root: tk.Wm,
    override: Callable[
        [
            str | None,
            Callable[[], None] | None,
            Callable[[str | None, Callable[[], None] | None], str],
        ],
        str,
    ],
) -> None:
    """Install protocol override.

    Basically, whenever root.protocol is called, instead it calls
    override with the name and function to bind and also passes the
    original root.protocol function
    """
    uninstall_protocol_override(root)
    original = root.protocol

    @wraps(original)
    def new_protocol(
        name: str | None = None,
        func: Callable[[], None] | None = None,
    ) -> str:
        return override(name, func, original)  # type: ignore[arg-type]

    root.protocol = new_protocol  # type: ignore[assignment]


def uninstall_protocol_override(root: tk.Wm) -> None:
    """Uninstall protocol override if it exists.

    If root.protocol has the __wrapped__ attribute, reset it
    to what it was originally
    """
    if hasattr(root.protocol, "__wrapped__"):
        root.protocol = root.protocol.__wrapped__  # type: ignore


class RunStatus(IntEnum):
    """Enum for trio run status.

    NO_TASK is when there is no task and not running
    TRIO_RUNNING is when there is task and running
    TRIO_RUNNING_CANCELED is when there is task and running but in the process
    of canceling task
    """

    TRIO_RUNNING_CANCELED = auto()
    TRIO_RUNNING_CANCELING = auto()
    TRIO_RUNNING = auto()
    TRIO_STARTING = auto()
    NO_TASK = auto()


def evil_does_trio_have_runner() -> bool:
    """Evil function to see if trio has a runner."""
    core = getattr(trio, "_core", None)
    if core is None:
        return False
    run = getattr(core, "_run", None)
    if run is None:
        return False
    global_run_context = getattr(run, "GLOBAL_RUN_CONTEXT", None)
    if global_run_context is None:
        return False
    return hasattr(global_run_context, "runner")


class TKMiscWmSubclass(tk.Misc, tk.Wm):
    """Subclass of both tkinter Misc and tkinter Wm."""

    __slots__ = ()


def is_tk_wm_and_misc_subclass(
    val: tk.Tk | tk.BaseWidget | tk.Wm | tk.Misc,
) -> TypeGuard[TKMiscWmSubclass]:
    """Return if value is an instance of both tk.Misc and tk.Wm."""
    return isinstance(val, tk.Misc) and isinstance(val, tk.Wm)


class TkTrioRunner:
    """Tk Trio Runner - Run Trio on top of Tkinter's main loop."""

    __slots__ = (
        "root",
        "call_tk_close",
        "nursery",
        "run_status",
        "recieved_loop_close_request",
        "installed_proto_override",
        "__weakref__",
    )

    def __new__(
        cls,
        root: tk.Wm | tk.Misc,
        *args: Any,
        **kwargs: Any,
    ) -> Self:
        """Either return new instance or get existing runner from root."""
        if not is_tk_wm_and_misc_subclass(root):
            raise ValueError("Must be subclass of both tk.Misc and tk.Wm")
        ref = getattr(root, "__trio__", None)

        if ref is not None:
            instance = ref()
            if instance is not None:
                if TYPE_CHECKING:
                    assert isinstance(instance, cls)
                assert instance.__class__.__name__ == cls.__name__
                return instance
        return super().__new__(cls)

    def __init__(
        self,
        root: tk.Wm | tk.Misc,
        restore_close: Callable[[], Any] | None = None,
    ) -> None:
        """Initialize trio runner."""
        if not is_tk_wm_and_misc_subclass(root):
            raise ValueError("Must be subclass of both tk.Misc and tk.Wm")
        if (
            hasattr(root, "__trio__")
            and getattr(root, "__trio__", lambda: None)() is self
        ):
            return
        self.root = root
        self.call_tk_close = restore_close or root.destroy

        self.nursery: trio.Nursery
        self.run_status = RunStatus.NO_TASK
        self.recieved_loop_close_request = False
        self.installed_proto_override = False

        with contextlib.suppress(AttributeError):
            root.__trio__ = weakref.ref(self)  # type: ignore[attr-defined]

    def schedule_task_threadsafe(
        self,
        function: Callable[..., Any],
        *args: Any,
    ) -> None:
        """Schedule task in Tkinter's event loop."""
        try:
            self.root.after_idle(function, *args)
        except RuntimeError as exc:
            debug(f"Exception scheduling task {function = }")
            # probably "main thread is not in main loop" error
            # mtTkinter is supposed to fix this sort of issue
            print_exception(exc)
            utils.extension_log_exception(exc)

            self.cancel_current_task()

    ##    def schedule_task_not_threadsafe(self, function: Callable[..., Any], *args: Any) -> None:
    ##        """Schedule task in Tkinter's event loop."""
    ##        try:
    ##            self.root.after_idle(function, *args)
    ##        except RuntimeError as exc:
    ##            debug(f"Exception scheduling task {function = }")
    ##            # probably "main thread is not in main loop" error
    ##            # mtTkinter is supposed to fix this sort of issue
    ##            print_exception(exc)
    ##
    ##            self.cancel_current_task()

    def cancel_current_task(self) -> None:
        """Cancel current task if one is running."""
        if self.run_status == RunStatus.NO_TASK:
            # No need to reschedule, is already closed.
            return

        if self.run_status == RunStatus.TRIO_STARTING:
            # Reschedule close for later, in process of starting.
            self.schedule_task_threadsafe(self.cancel_current_task)
            return

        if self.run_status == RunStatus.TRIO_RUNNING_CANCELING:
            # Already scheduled to cancel
            return

        self.run_status = RunStatus.TRIO_RUNNING_CANCELING
        try:
            self.nursery.cancel_scope.cancel()
        except RuntimeError as exc:
            # probably "must be called from async context" error
            # because the exception that triggered this was from
            # a start group tick failing because of start_soon
            # not running from main thread because thread lock shenanigans
            print_exception(exc)
            utils.extension_log_exception(exc)

            # We can't even show an error properly because of the same
            # issue!
            try:
                self.show_irrecoverable_error_stopping(
                    "".join(format_exception(exc)),
                )
            except RuntimeError as exc:
                print_exception(exc)
                utils.extension_log_exception(exc)
        else:
            self.run_status = RunStatus.TRIO_RUNNING_CANCELED

    def _start_async_task(
        self,
        function: Callable[[], Awaitable[Any]],
    ) -> None:
        """Run async task in new nursery."""
        if evil_does_trio_have_runner():
            self.show_warning_trio_already_running()
            return

        @trio.lowlevel.disable_ki_protection
        async def run_nursery() -> None:
            """Run nursery."""
            assert self.run_status == RunStatus.TRIO_STARTING
            async with trio.open_nursery(
                strict_exception_groups=True,
            ) as nursery:
                self.nursery = nursery
                self.run_status = RunStatus.TRIO_RUNNING
                self.nursery.start_soon(function)
            return

        def done_callback(outcome: Outcome[None]) -> None:
            """Handle when trio is done running."""
            assert self.run_status in {
                RunStatus.TRIO_RUNNING_CANCELED,
                RunStatus.TRIO_RUNNING,
            }
            self.run_status = RunStatus.NO_TASK
            del self.nursery
            try:
                outcome.unwrap()
            except ExceptionGroup as exc:
                print_exception(exc)
                utils.extension_log_exception(exc)

        if self.run_status != RunStatus.NO_TASK:
            raise RuntimeError(
                "Cannot run more than one trio instance at once.",
            )

        self.run_status = RunStatus.TRIO_STARTING

        trio.lowlevel.start_guest_run(
            run_nursery,
            done_callback=done_callback,
            run_sync_soon_threadsafe=self.schedule_task_threadsafe,
            host_uses_signal_set_wakeup_fd=False,
            restrict_keyboard_interrupt_to_checkpoints=True,
            strict_exception_groups=True,
        )

    def get_del_window_proto(
        self,
        new_protocol: Callable[[], None] | None,
    ) -> Callable[[], None]:
        """Create new WM_DELETE_WINDOW protocol to shut down trio properly."""

        def shutdown_then_call() -> None:
            # If this is first time, withdraw root
            if not self.recieved_loop_close_request:
                self.root.withdraw()
            self.recieved_loop_close_request = True
            # If a task is still running
            if self.run_status == RunStatus.TRIO_RUNNING_CANCELING:
                # Error happened and cancel failed
                # Hopefully will be ok because closing host loop
                self.run_status = RunStatus.NO_TASK
                debug(
                    "Changing run status Canceling -> No Task, error happened during canceling.",
                )
            if self.run_status == RunStatus.TRIO_RUNNING:
                # Cancel it if not already canceled
                self.cancel_current_task()
            if self.run_status != RunStatus.NO_TASK:
                # Rerun this function again in the future until no task running.
                self.schedule_task_threadsafe(shutdown_then_call)
                return None
            # No more tasks
            # Make sure to uninstall override or IDLE gets mad
            uninstall_protocol_override(self.root)
            self.installed_proto_override = False
            # If close function exists, call it.
            if new_protocol is not None:
                return new_protocol()
            return None

        return shutdown_then_call

    def host_loop_close_override(
        self,
        name: str | None,
        new_protocol: Callable[[], None] | None,
        original_bind_proto: Callable[
            [str | None, Callable[[], None] | None],
            str,
        ],
    ) -> str:
        """Override for new protocols on tkinter root.

        Catch new protocols for `WM_DELETE_WINDOW` and replace them
        with our own that will cancel trio run and make sure it closes, and
        then call the new protocol
        """
        if name != "WM_DELETE_WINDOW":
            # Not important to us
            return original_bind_proto(name, new_protocol)
        return original_bind_proto(
            name,
            self.get_del_window_proto(new_protocol),
        )

    def show_warning_trio_already_running(self) -> None:
        """Show warning that trio is already running."""
        messagebox.showerror(
            title="Error: Trio is already running",
            message="Trio is already running from somewhere else, "
            "please try again later.",
            parent=self.root,
        )

    def show_irrecoverable_error_stopping(
        self,
        extra_message: str = "",
    ) -> None:
        """Show warning that previous trio run needs to stop."""
        messagebox.showwarning(
            title="Irrecoverable Error While Stopping",
            message=(
                "Encountered an irrecoverable error stopping a previous "
                "Trio run. Please restart IDLE and report this on Github at "
                "https://github.com/CoolCat467/idlemypyextension/issues"
                f"{extra_message}"
            ),
            parent=self.root,
        )

    def no_start_trio_is_stopping(self) -> None:
        """Show warning that previous trio run needs to stop."""
        messagebox.showwarning(
            title="Warning: Trio is stopping a previous run",
            message="Trio is in the process of stopping, "
            "please try again later.",
            parent=self.root,
        )

    def __call__(self, function: Callable[[], Awaitable[Any]]) -> None:
        """Schedule async function for the future."""
        # If host loop requested to close, do not run any more tasks.
        if self.recieved_loop_close_request:
            return None

        if self.run_status == RunStatus.TRIO_RUNNING_CANCELING:
            # Task status is in a limbo state where it tried to
            # cancel the nursery but something went wrong.
            # This should not happen with mtTkinter's fixes, but
            # it might still be a possibility.
            self.show_irrecoverable_error_stopping()
            return None

        if self.run_status == RunStatus.TRIO_RUNNING_CANCELED:
            ### Task run is in the process of stopping
            ### self.no_start_trio_is_stopping()
            # Reschedule starting task
            self.schedule_task_threadsafe(self.__call__, function)
            return None

        if self.run_status == RunStatus.TRIO_STARTING:
            # Reschedule starting task
            self.schedule_task_threadsafe(self.__call__, function)
            return None

        # If there is a task running
        if self.run_status == RunStatus.TRIO_RUNNING:
            self.nursery.start_soon(function)
            return None

        if self.run_status != RunStatus.NO_TASK:
            raise RuntimeError(f"Invalid run status {self.run_status!r}")

        # If we have not installed the protocol override,
        if not self.installed_proto_override:
            # Install it
            install_protocol_override(self.root, self.host_loop_close_override)
            self.installed_proto_override = True
            # Make sure new handler is now applied
            self.root.protocol("WM_DELETE_WINDOW", self.call_tk_close)

        # Start running new task
        return self._start_async_task(function)


def run() -> None:
    """Run test of module."""

    # A tiny Trio program
    async def trio_test_program(run: int) -> str:
        for i in range(5):
            print(f"Hello from Trio! ({i}) ({run = })")
            # This is inside Trio, so we have to use Trio APIs
            try:
                await trio.sleep(1)
            except trio.Cancelled:
                print(f"Run {run} canceled.")
                raise
        return "trio done!"

    root = tk.Tk(className="Trio Test")

    def trigger_trio_runs() -> None:
        trio_run = TkTrioRunner(root)
        trio_run(partial(trio_test_program, 0))
        trio_run_2 = TkTrioRunner(root)
        trio_run_2(partial(trio_test_program, 1))
        trio_run_2(partial(trio_test_program, 2))
        trio_run(partial(trio_test_program, 3))

    # Wait a tiny bit so main loop is running from main thread
    root.after(10, trigger_trio_runs)
    print("synchronous after trio run start")
    # If testing without IDLE, uncomment the line below:
    # root.mainloop()


if __name__ == "__main__":
    print(f"{__title__}\nProgrammed by {__author__}.\n")
    run()



"""moduleguard - Guard import(s) from IDLE interfering with system path."""

# Programmed by CoolCat467

from __future__ import annotations

# moduleguard - Guard import(s) from IDLE interfering with system path.
# Copyright (C) 2023  CoolCat467
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__title__ = "moduleguard"
__author__ = "CoolCat467"
__license__ = "GNU General Public License Version 3"
__version__ = "0.0.0"

import json
import os
import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterable
    from types import TracebackType

    from typing_extensions import Self


def does_path_interfere(
    path: str,
    modules: set[str],
    *,
    ignorepaths: set[str] | None = None,
) -> bool:
    """Return True if path contains something that could interfere.

    Arguments:
    ---------
      path: Path to check.
      modules: Set of module names we want to make sure won't interfere.
      ignorepaths: Set of paths already checked and known won't interfere.

    This is a recursive function that mutates ignorepaths in each iteration.
    It is advised against passing the ignorepaths argument.

    """
    # Setup ignorepaths set when user calls this
    if ignorepaths is None:
        ignorepaths = set()

    for dirpath, dirnames, filenames in os.walk(path, topdown=True):
        # Remove directories we have already visited
        for dirname in tuple(dirnames):
            if os.path.join(dirpath, dirname) in ignorepaths:
                dirnames.remove(dirname)

        for filename in filenames:
            # Skip filenames with no extension
            if os.path.extsep not in filename:
                continue
            # If filename is module name, that's bad
            if filename.split(os.path.extsep, 1)[0] in modules:
                return True

        # If this is a module internal directory
        if "__init__.py" in filenames:
            # Get the next level up's name and the module name
            level_up, dir_module_name = os.path.split(path)
            # If module name matches, bad
            if dir_module_name in modules:
                return True
            # Otherwise, everything is fine and we have checked this folder.
            # No need for next iteration to do so again
            ignorepaths.add(path)
            # Need to check level up now
            return does_path_interfere(
                level_up,
                modules,
                ignorepaths=ignorepaths,
            )
    # If nothing matched we should be fine
    return False


class ImportGuardContextManager:
    """Guard imports against user packages from idle's sys.path manipulation."""

    __slots__ = ("modules", "original")

    def __init__(self, modules: set[str]) -> None:
        """Initialize modules set."""
        self.modules = modules
        self.original: list[str] = []

        bad_modules = ", ".join(modules & set(sys.builtin_module_names))
        if bad_modules:
            raise ValueError(
                f"Cannot guard following builtin modules: {bad_modules}",
            )

    def __repr__(self) -> str:
        """Return representation of self."""
        return f"{self.__class__.__name__}({self.modules!r})"

    def __enter__(self) -> Self:
        """Modify sys.path to remove interference."""
        # Get deep copy
        self.original = json.loads(json.dumps(sys.path))

        # Remove blanks
        index = 0
        while index < len(sys.path):
            if not sys.path[index]:
                path = sys.path.pop(index)
                # print(f"[DEBUG] popped sys.{path = }")
                continue
            index += 1
        # print(f"[DEBUG] {sys.path = }")

        if "idlelib" not in sys.modules:
            # We are in before IDLE, we should be safe
            return self

        # Remove conflict(s)
        # First, find where the end of idle's manipulation is
        max_read = 0
        for max_read, path in enumerate(sys.path):  # noqa: B007
            if path.startswith(sys.exec_prefix):
                break
        max_read = min(max_read, 2)

        index = 0
        while index < max_read:
            if does_path_interfere(sys.path[index], self.modules):
                path = sys.path.pop(index)
                # print(f"[DEBUG] popped sys.{path = }")
                max_read -= 1
                continue
            index += 1

        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Restore sys.path to what it was before we changed it."""
        sys.path.clear()
        for item in self.original:
            sys.path.append(item)


def guard_import(module_name: str) -> ImportGuardContextManager:
    """Guard an import against user packages from idle's sys.path manipulation."""
    return ImportGuardContextManager({module_name})


def guard_imports(module_names: Iterable[str]) -> ImportGuardContextManager:
    """Guard specified imports against user packages from idle's sys.path manipulation."""
    return ImportGuardContextManager(set(module_names))


if __name__ == "__main__":
    print(f"{__title__}\nProgrammed by {__author__}.\n")



"""Type Check IDLE Extension."""

# Programmed by CoolCat467

from __future__ import annotations

# Copyright (C) 2023  CoolCat467
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__title__ = "idlemypyextension"
__author__ = "CoolCat467"
__license__ = "GNU General Public License Version 3"
__version__ = "1.0.4"


from idlemypyextension import utils
from idlemypyextension.extension import idlemypyextension as idlemypyextension


def check_installed() -> bool:
    """Make sure extension installed."""
    return utils.check_installed(__title__, __version__, idlemypyextension)


utils.set_title(__title__)
idlemypyextension.reload()


if __name__ == "__main__":
    print(f"{__title__} v{__version__}\nProgrammed by {__author__}.\n")
    check_installed()
